#include <sys/types.h>
#include <ctype.h>
#include <stdio.h>
#include <time.h>
#include <conio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>
#include "ffind.h"

#include "utime.h"

#include "com.h"
#include "xfer.h"
#include "zmodem.h"
#include "keybd.h"
#include "sbuf.h"
#include "sched.h"
#include "externs.h"
#include "prototyp.h"

/* #pragma loop_opt(off) */

static void SendACK(void);
static void SendNAK(void);
static void getblock(void);

/*--------------------------------------------------------------------------*/
/* LOCALS                                                                   */
/*--------------------------------------------------------------------------*/
static long block_number;
static long base_block;
static char *stat_msg;
static int sliding;
static int block_size;
static int do_chksum;
static int errs;
static int real_errs;
static long fsize1;
static int fsize2;
static char *buffer;
static FILE *infile;
static char *_fpath;
static struct zero_block *header;
static char final_name[80];
static int first_block;                 /* 1 = first block/Matrix packet */


 /*static int netmail;*//* PLF Tue  10-31-1989  18:53:47 */
static int recv_ackless;
static int did_nak;
static long filesecs;

#define two_s "%s%s"



/*--------------------------------------------------------------------------*/
/* SEND ACK                                                                 */
/*--------------------------------------------------------------------------*/
static void SendACK()
{
    if ((!recv_ackless) || (block_number == 0)) {
        SENDBYTE(ACK);
        if (sliding) {
            SENDBYTE((unsigned char) block_number);
            SENDBYTE((unsigned char) (~(block_number)));
        }

        if (block_number <= fsize1) {
            /* No-nonsense status line.  Don't waste time with frills... */
            Lprintf("%lu\r", block_number);
        }
    }

    else if (((!(block_number & 0x001F)) && (block_number < fsize1))
        || (block_number == fsize1)) {
        /* In Overdrive only put out last block and n/32 */
        Lprintf("%lu*\r", block_number);
    }

    errs = 0;
}                                       /* SendACK */



/*--------------------------------------------------------------------------*/
/* SEND NAK                                                                 */
/*--------------------------------------------------------------------------*/
static void SendNAK()
{
    int i;
    long t1;

    errs++;
    real_errs++;
    if (errs > 6) {
        stat_msg = FUBAR_msg;
        return;
    }

    if (++did_nak > 8) {
        recv_ackless = 0;
    }

    CLEAR_INBOUND();

    /*--------------------------------------------------------------------*/
    /* Let the connection cool its heels.                                 */
    /*--------------------------------------------------------------------*/
    if (!recv_ackless) {
        /* Set a timer for waiting */
        t1 = timerset(3000);
        if ((base_block != block_number) || (errs > 1))
            do {
                i = TIMED_READ(1);
                if (!CARRIER())
                    return;
                if (timeup(t1))
                    break;
            }
            while (i >= 0);
    }

    if (block_number > base_block)
        SENDBYTE(NAK);
    else {
        if ((errs < 5) && (!do_chksum))
            SENDBYTE('C');
        else {
            do_chksum = 1;
            SENDBYTE(NAK);
        }
    }

    if (sliding) {
        SENDBYTE((unsigned char) block_number);
        SENDBYTE((unsigned char) (~(block_number)));
    }

    if (block_number <= fsize1) {

        /* No time for anything but the bare essentials... */
        Lprintf("%lu\r", block_number);
    }
}                                       /* SendNAK */

/*--------------------------------------------------------------------------*/
/* GET BLOCK                                                                */
/*--------------------------------------------------------------------------*/
static void getblock()
{
    REGISTER int i;
    REGISTER char *sptr;

    unsigned int crc;
    int in_char;
    int is_resend;
    int blockerr;
    unsigned char chksum;

    blockerr = is_resend = 0;
    chksum = '\0';


    /*--------------------------------------------------------------------*/
    /* HEADER: Block number                                               */
    /*--------------------------------------------------------------------*/
    in_char = TIMED_READ(5);
    if (in_char != ((int) (block_number) & 0xff)) {
        if (in_char < (int) block_number)
            is_resend = 1;
        else if ((block_number) || (in_char != 1)) {
            blockerr++;
            stat_msg = SYNC_msg;
        }
        else
            block_number = 1;
    }

    /*--------------------------------------------------------------------*/
    /* HEADER: Complement of the block number                             */
    /*--------------------------------------------------------------------*/
    i = TIMED_READ(5);
    if ((i & 0xff) != ((~in_char) & 0xff)) {
        blockerr++;
        stat_msg = CMPL_msg;
    }

    /*--------------------------------------------------------------------*/
    /* DATA                                                               */
    /*--------------------------------------------------------------------*/
    for (sptr = buffer, i = 0; i < block_size; i++, sptr++) {
        in_char = TIMED_READ(5);
        if (in_char < 0) {              /* We get EOF for timeout or
                                         * carrier */
            if (CARRIER()) {              /* If we have carrier it was
                                         * timeout */
                SendNAK();
                stat_msg = TIME_msg;
            }
            return;                     /* return in either case */
        }
        sptr[0] = (char) in_char;
    }

    /*--------------------------------------------------------------------*/
    /* CHECK                                                              */
    /*--------------------------------------------------------------------*/
    if (do_chksum) {
        for (sptr = buffer, i = 0; i < block_size; i++, sptr++)
            chksum += sptr[0];

        if (TIMED_READ(5) != chksum) {
            stat_msg = CHK_msg;
            blockerr++;
        }
    }
    else {
        unsigned int lsb, msb;

        for (sptr = buffer, i = crc = 0; i < block_size; i++, sptr++)
            crc = xcrc(crc, (byte) sptr[0]);

        msb = TIMED_READ(3);
        lsb = TIMED_READ(3);
        if ((lsb < 0) || (msb < 0)) {
            stat_msg = SHRT_msg;
            if (!block_number)
                sliding = 0;
            blockerr++;
        }
        else if (((msb << 8) | lsb) != crc) {
            stat_msg = CRC_msg;
            blockerr++;
        }
    }

    if (blockerr)
        SendNAK();
    else {
        SendACK();
        if (is_resend)
            return;

        if (block_number) {
            int ok;
            if (block_number < fsize1)
                ok = (1==fwrite(buffer, block_size, 1, infile));
            else if (block_number == fsize1)
                ok = (1==fwrite(buffer, fsize2, 1, infile));
            if(!ok)
                got_error(WRITE_msg, final_name);
            else
                errno = 0;
        }
        else {
            if (!first_block) {
                sptr = &(header->name[0]);

                for (i = 0; ((sptr[i]) && (i < 17)); i++)
                    if (sptr[i] <= ' ')
                        sptr[i] = '\0';

                if (sptr[0]) {
                    sprintf(final_name, two_s, _fpath, sptr);
                    sptr = final_name;
                    fancy_str(sptr);

                    i = strlen(final_name) - 1;

                    Lprintf("Receiving %s, ", sptr);
                }
                else
                    status_line("!Grunged hdr");
            }

            if (header->size) {
                fsize1 = (long) (header->size / 128L);
                if (fsize2 = (int) (header->size % 128))
                    ++fsize1;
                else
                    fsize2 = 128;
                Lprintf("%ld blocks", fsize1);
            }
            if (header->time) {
                filesecs = header->time;
            }
            else {
                filesecs = -1;
            }

            if (header->moi[0]) {
                Lprintf(" from %s", header->moi);
            }
            if ((cur_baud >= 9600) && header->noacks && !no_overdrive) {
                recv_ackless = 1;
            }
            Lprintf("\n");
        }
        block_number++;
    }
}                                       /* getblock */

/*--------------------------------------------------------------------------*/
/* GENERAL PURPOSE FILE RECEPTION ROUTINE                                   */
/*--------------------------------------------------------------------------*/
char *receive_file(char *fpath, char *fname, char protocol)
{
    char tmpname[80];
    int in_char;
    struct utimbuf times;
    long t1;

    did_nak = 0;
    filesecs = -1;
    XON_DISABLE();
    first_block = 0;

    CLEAR_IOERR();
    fsize1 = 65535;
    if (fname)
        for (in_char = 0; fname[in_char]; in_char++)
            if ((fname[in_char] == '*') || (fname[in_char] == '?'))
                fname[0] = '\0';

    _fpath = fpath;
    sliding = 1;
    base_block =
        do_chksum =
        errs = 0;
    block_size = 128;

    switch (protocol) {
        case 'X':
            base_block = 1;
            sliding = 0;
            break;
        case 'Y':
            base_block = 1;
            sliding = 0;
            block_size = 1024;
            break;
        case 'S':
            break;
        case 'T':
            break;
        case 'M':
            base_block = 1;
            sliding = 0;
            break;
        default:
            status_line("!Protocol??");
            return NULL;
    }

    block_number = base_block;

    sprintf(tmpname, "%s_TMP%02x_.$$$", fpath, task_num);
    sprintf(final_name, two_s, _fpath, (fname && fname[0]) ? fname : "UNKNOWN.$$$");

    infile = fopen(tmpname, write_binary);
    if (!infile){
        got_error(CREATE_msg, tmpname);
        return (NULL);
    }
    if (isatty(fileno(infile))) {
        errno = 1;
        got_error(DEVICE_msg, tmpname);
        fclose(infile);
        return (NULL);
    }

    buffer = Txbuf;
    header = (struct zero_block *) buffer;
    if (!buffer) {
        status_line("!MEM:ulbuf");
        return (NULL);
    }

    stat_msg = NULL;
    throughput(0, 0L);

/*  if(protocol != 'S')
        SendNAK();   *PLF Tue  11-21-1989  11:39:37 */

    t1 = timerset(300);
    real_errs = 0;

loop_top:

    if (got_ESC()) {
        stat_msg = KBD_msg;
        goto fubar;
    }

    in_char = TIMED_READ(4);

    CLEAR_IOERR();                      /* PLF Sat  11-18-1989  23:30:30 */
    switch (in_char) {
        case SOH:
            /*--------------------------------------------*/
            /* SOH: 128 byte block header                 */
            /*--------------------------------------------*/
            block_size = 128;
            getblock();
            t1 = timerset(300);
            break;

        case STX:
            /*--------------------------------------------*/
            /* STX: YModem 1k block header                */
            /*--------------------------------------------*/
            block_size = 1024;
            getblock();
            t1 = timerset(300);
            break;

        case SYN:
            /*--------------------------------------------*/
            /* SYN: TeLink Zero block header              */
            /*--------------------------------------------*/
            do_chksum = 1;
            getblock();
            do_chksum = 0;
            t1 = timerset(300);
            break;

        case CAN:
            /*--------------------------------------------*/
            /* CAN                                        */
            /*--------------------------------------------*/
            if (TIMED_READ(2) == CAN) {
                stat_msg = CAN_msg;
                goto fubar;
            }
            t1 = timerset(300);
            break;

        case EOT:
            /*--------------------------------------------*/
            /* EOT                                        */
            /*--------------------------------------------*/
            stat_msg = EOT_msg;
            t1 = timerset(20);
            while (!timeup(t1)) {
                if( CHAR_AVAIL() )
                    TIMED_READ(0);
                else
                    time_release();
            }
            if (block_number)
                goto done;
            else
                goto fubar;

        default:
            /*--------------------------------------------*/
            /* Default                                    */
            /*--------------------------------------------*/
            if (in_char > 0) {
/*                Lprintf("??? [%Xh, blk %d] \r", in_char, block_number); */
            }
            else if (!CARRIER())          /* Was it a timeout or carrier
                                         * loss? */
                goto lost;              /* Carrier, get out of here          */

            /*
             * Debris is tolerable for up to the amount of time we are
             * willing to wait for a start of a block.  Right now it is 10
             * seconds.
             */

            /*
             * If may_be_seadog is set, then we know we should get an SOH
             * pretty soon
             */
            if(timeup(t1))
                SendNAK();
            break;
    }                                   /* switch */

    if (errs > 14) {
        stat_msg = FUBAR_msg;
        goto fubar;
    }

    if (stat_msg) {
        Lputs("\r                         \r");
        Lputs(stat_msg);
        stat_msg = NULL;
    }
    goto loop_top;

lost:

    stat_msg = CARRIER_msg;

fubar:

    if (infile) {
        if(fclose(infile))
            got_error(CLOSE_msg, tmpname);
        if(remove(tmpname))
            got_error(UNLINK_msg, tmpname);
    }

    CLEAR_OUTBOUND();

    if (stat_msg != EOT_msg) {
        for (in_char = 0; in_char < 5; in_char++)
            SENDBYTE(CAN);

        status_line("!File not received: %s", stat_msg);
    }

    CLEAR_INBOUND();
    return NULL;


done:

    recv_ackless = 0;

    SendACK();

    if (infile) {
        int i, j, k;
        FFIND *ff; /*PLF Fri  02-16-1990  15:37:45 */

        if(fclose(infile))
            got_error(CLOSE_msg, tmpname);

        i = strlen(tmpname) - 1;
        j = strlen(final_name) - 1;

        if (tmpname[i] == '.')
            tmpname[i] = '\0';
        if (final_name[j] == '.') {
            final_name[j] = '\0';
            --j;
        }

        i = 0;
        k = is_arcmail(final_name, j);
        if ((!overwrite) || k) {
            while (rename(tmpname, final_name)) {
                if (isdigit(final_name[j]))
                    final_name[j]++;
                else
                    final_name[j] = '0';
                if (!isdigit(final_name[j]))
                    return (tmpname);
                i = 1;
            }
            CLEAR_IOERR();
        }
        else {
            unlink(final_name);
            rename(tmpname, final_name);
        }
        if (i) {
            status_line("+Dupe file renamed: %s", final_name);
        }

        if (filesecs != -1) {
            tzset();
            times.modtime = filesecs + timezone;
            times.actime = filesecs + timezone;
            utime(final_name, &times);
        }

        if (NULL != (ff=FindOpen(final_name, 0)))
        {
            extern char *offsets;

            if (real_errs > 4)
                status_line("+Corrected %d errors in %d blocks",
                    real_errs, fsize1);

            throughput(1, ff->ff_fsize);

            status_line("=UL-%c %s", protocol, final_name);

            update_files(0);

            if (prm.k_free && zfree(final_name) < (long)prm.k_free*1000L)
            {
              logit("!Not enough free space for UL, deleted");
              unlink(final_name);
              ul_no_space=TRUE;

              CLEAR_IOERR();
              FindClose(ff);
              return NULL;
            }
            else strcpy(final_name, ff->ff_name);

            CLEAR_IOERR();
            FindClose(ff);

            return final_name;          /* signal what file we got */
        }
    }
    CLEAR_IOERR();
    return NULL;

}

/* END OF FILE: recv_sl.c */
